import re

class units:
    unit_categories = {
            'distance': ['m', 'ft', 'in', 'mi'],
            'time': ['s', 'min', 'hour', 'hr', 'h', 'day', 'd'],
            'mass': ['kg', 'g', 'lb', 'oz'],
            'temperature': ['K', 'C', 'F'],
            'current': ['A'],
            'amount': ['molecules','molecule','molec', 'mQ', 'mol'],  # longer units first
            'volume': ['L', 'gal'],
            'energy': ['J','cal','Cal'],
            'frequency': ['Hz'],
            'force': ['N'],
            'pressure': ['Pa','bar','mmHg','Torr','atm','psi'],
            'power': ['W']
        }
    
    si_prefixes = {
            'Y': 1e24, 'Z': 1e21, 'E': 1e18, 'P': 1e15,
            'T': 1e12, 'G': 1e9, 'M': 1e6, 'k': 1e3,
            'h': 1e2, 'da': 1e1, 'd': 1e-1, 'c': 1e-2,
            'm': 1e-3, 'µ': 1e-6, 'n': 1e-9, 'p': 1e-12,
            'f': 1e-15, 'a': 1e-18, 'z': 1e-21, 'y': 1e-24
        }
    
    categories_si = {
            'distance': 'm',
            'time': 's',
            'mass': 'kg',
            'temperature': 'K',
            'current': 'A',
            'amount': 'mol',
            'volume': 'm3',
            'energy': 'kg m2 s-2',
            'frequency': 's-1',
            'force': 'kg m s-2',
            'pressure': 'kg m-1 s-2',
            'power': 'kg m2 s-3'
        }

    to_si_factors = {
            'ft': 0.3048,     # foot to meter
            'in': 0.0254,     # inch to meter
            'mi': 1609.34,    # mile to meter
            'min': 60,        # minute to second
            'h': 3600,        # hour to second
            'hr': 3600,       # hour to second
            'hour': 3600,       # hour to second
            'd': 86400,       # day to second
            'day': 86400,     # day to second
            'g': 1e-3,        # gram to kilogram
            'lb': 0.453592,   # pound to kilogram
            'oz': 0.0283495,  # ounce to kilogram
            # 'C': 'x+273.15',           # Celsius to Kelvin; note: requires offset
            # 'F': '(x-32)*5/9',         # Fahrenheit to Kelvin; note: requires offset and scaling
            'molec':  1 / 6.02214076e23,  # molecules to mole
            'mQ':  1 / 6.02214076e23,  # molecules to mole
            'molecules': 1 / 6.02214076e23,  # molecules to mole
            'molecule':1 / 6.02214076e23,
            'L': 1e-3,        # liter to cubic meter
            'gal': 0.00378541, # US gallon to cubic meter
            'cal': 4.184,     # calorie to joule
            'Cal': 4184,      # Calorie (kilocalorie) to joule
            'bar': 1e5,       # bar to pascal
            'mmHg': 133.322,  # millimeter of mercury to pascal
            'Torr': 133.322,  # torr to pascal
            'atm': 101325,    # standard atmosphere to pascal
            'psi': 6894.76,   # pound-force per square inch to pascal
        }


    def __init__(self, units_str):
        self.units_str = units_str
        self.to_si_factor = 1
        self.unit_coefs = self._parse_units()
        self.units_str_si = self._construct_si_units_str()
        self.unit_type = self._determine_unit_type()
        self.tex = self._format_tex()

    def _parse_units(self, str_to_parse=None):
        str_to_parse = self.units_str if str_to_parse is None else str_to_parse    

        unit_exponents = {
            'mass': 0,
            'distance': 0,
            'time': 0,
            'temperature': 0,
            'current': 0,
            'amount': 0
        }

        unit_list = str_to_parse.split(' ')
        for unit_exp in unit_list:
            # Extract the unit and exponent if present
            match = re.match(r'([a-zA-Zµ]*)([-0-9]*)', unit_exp)
            if match:
                unit, exp = match.groups()

                # Attempt to match the entire unit first
                unit_found = False
                for category, units in self.unit_categories.items():
                    for base_unit in units:
                        if unit == base_unit:
                            unit_found = True
                            prefix = ''
                            break

                    if unit_found:
                        break

                # If no match was found, try to split into a prefix and a base unit and try again
                if not unit_found and len(unit) > 1:
                    for category, units in self.unit_categories.items():
                        for base_unit in units:
                            if unit[1:] == base_unit:
                                unit_found = True
                                prefix = unit[0]
                                break

                        if unit_found:
                            break

                if unit_found:
                    # Check for exponent
                    exponent = int(exp) if exp else 1
                    
                    if category in unit_exponents:  # if category is a key in unit_exponents
                        unit_exponents[category] += exponent
                    else:  # recursively parse units string from categories_si
                        si_unit_str = self.categories_si[category]
                        si_unit_str = f"{si_unit_str}"
                        si_unit_exponents = self._parse_units(si_unit_str)
                        for si_category, si_exponent in si_unit_exponents.items():
                            unit_exponents[si_category] += si_exponent * exponent

                    # Adjust self.to_si_factor with conversion factor and prefix
                    conv_factor = self.to_si_factors.get(base_unit, 1)
                    prefix_factor = self.si_prefixes.get(prefix, 1)
                    self.to_si_factor *= (conv_factor * prefix_factor) ** exponent
        return unit_exponents




    def _construct_si_units_str(self):
        units_str_si_parts = []
        for category, exponent in self.unit_coefs.items():
            if exponent != 0:
                si_unit = self.categories_si[category]
                units_str_si_part = f"{si_unit}{exponent if exponent != 1 else ''}"
                units_str_si_parts.append(units_str_si_part)
        return ' '.join(units_str_si_parts)

    def _determine_unit_type(self):
        for unit_type, si_units_str in self.categories_si.items():
            if self.units_str_si == si_units_str:
                return unit_type
        return None

    def __str__(self):
        return self.units_str
    
    def convert_to(self, output_unit_str):
        output_unit = units(output_unit_str)
        
        factor = self.to_si_factor / output_unit.to_si_factor

        if output_unit.unit_type is not self.unit_type:
            factor = None

        return factor
    
    def _format_tex(self):
        numerator_units = []
        denom_units = []
        for u in self.units_str.split(' '):
            pattern = r'([a-zA-Z]+)(-?\d+)?'
            match = re.match(pattern, u)
            if match:
                unit = match.group(1)
                power = match.group(2)

                if power:
                    power = int(power)
                    
                    if abs(power) > 1:
                        unit = unit + '^{' + str(abs(power)) + '}'
                    
                    if power < 0:
                        denom_units.append(unit)
                    else:
                        numerator_units.append(unit)
                else:
                    numerator_units.append(unit)

        # Handle the case where there's no denominator
        if denom_units:
            return '\\frac{' + '\\,'.join(numerator_units) + '}{' + '\\,'.join(denom_units) + '}'
        else:
            return '\\,'.join(numerator_units)

        return ''
    

def main():
    test_unit = units("mi3 day-2 s-1")
    print(test_unit.unit_coefs)
    print(test_unit.units_str_si)
    print(test_unit.unit_type)
    print(test_unit.to_si_factor)
    print(test_unit.tex)

if __name__ == "__main__":
    main()


from setuptools import setup, find_packages

setup(
    name="dp_chem",
    version="1.2",
    packages=find_packages(),
    description="A package for handling chemistry concepts and data",
    author="Demetrios Pagonis",
    author_email="demetriospagonis@weber.edu",
    include_package_data=True,
    install_requires=[
        'numpy',
        'pandas',
        'jinja2',
        'scipy',
        'matplotlib',
        'xarray',
        'requests',
        'cartopy',
        'shapely',
        'geopandas',
    ],
)


import re
from itertools import permutations

from .molecule import molecule




class reaction:
    """
    Class representing a chemical reaction.

    Attributes:
        isbalanced (bool): Boolean result of checking charge and mass balance.
        plaintext (str): String with plain text representation of the reaction.
        plaintext_answers (str): All permutations of order in plaintext.
        Q_eqn (str): String with plain text representation of the reaction coefficient.
        Q_eqn_answers (str): All permutations of order in Q_eqn.
        Q_tex (str): TeX representation of the reaction quotient.
        reactants (list): Lists of molecule objects representing the reactants.
        products (list): Lists of molecule objects representing the products.
        reversible (bool): Boolean that determines what type of arrow goes in self.tex.
        tex (str): String with reaction as TeX.

    Methods:
        setconcentration(formula, conc): Function to set the concentration of a given molecule. Pass the formula as a string.
    """


    def __init__(self, reactants, products=None,reversible=True,k=None):

        # Introducing flag to track if initial setup is in progress
        self._initializing = True
        
        if "=" in reactants and products is None: #entered the whole reaction directly as a string
            reactants, products = reactants.split("=")

        self.reactants = self._parse_and_validate_input(reactants)
        self.products = self._parse_and_validate_input(products)
        self.reversible = reversible
        self.k=k
        # End of initialization
        self._initializing = False
        self._update()  # Call update once at the end of initialization

    @property
    def reactants(self):
        return self._reactants

    @reactants.setter
    def reactants(self, new_value):
        self._reactants = self._parse_and_validate_input(new_value)
        if not self._initializing and hasattr(self, '_products'):  # Check if _products exists and we are not in initialization phase
            self._update()

    @property
    def products(self):
        return self._products

    @products.setter
    def products(self, new_value):
        self._products = self._parse_and_validate_input(new_value)
        if not self._initializing and hasattr(self, '_reactants'):  # Check if _reactants exists and we are not in initialization phase
            self._update()

    def _update(self):
        self.plaintext, self.plaintext_answers = self._generate_rxn_plain_text()
        self.tex = self._generate_rxn_tex()
        self.isbalanced = self._check_balance()
        self.Q_eqn, self.Q_eqn_answers, self.Q_tex, self.Q = self._generate_rxn_quotient()
        self.rate_eqn, self.rate_eqn_answers, self.rate_eqn_tex, self.rate = self._generate_rate_expression()
        self.simple_html = self._generate_simple_html()

    def _parse_and_validate_input(self, input_val):
        if isinstance(input_val, str):
            return self._parse_input_str(input_val)
        elif isinstance(input_val, molecule):
            return [input_val]
        elif isinstance(input_val, list) and all(isinstance(i, str) for i in input_val):
            return[self._parse_compound_str(i) for i in input_val]
        elif isinstance(input_val, list) and all(isinstance(i, molecule) for i in input_val):
            return input_val
        else:
            s = f'list of {type(input_val[0])}' if isinstance(input_val, list) else type(input_val)
            raise TypeError(f'{self.__class__.__name__} was passed a {s} as input. input = {input_val}')

    def _parse_compound_str(self,s):
        s_in = s
        s = s.strip()
        phase_re = re.compile(r'\(([a-z\s]+)\)$')
        coef_formula_re = re.compile(r'(\d+)?([A-Za-z\(\)\d\+\-]+)')

        phase_match = phase_re.search(s)

        if phase_match:
            phase = phase_match.group(1)
            s = s[:phase_match.start()]  # remove the phase from the string
        else:
            phase = None  # if no phase is found
        
        coef_formula_match = coef_formula_re.match(s.strip())  # strip any whitespace

        if not coef_formula_match:
            raise ValueError(f"Invalid compound format: {s_in}")

        coef = int(coef_formula_match.group(1)) if coef_formula_match.group(1) else 1
        formula = coef_formula_match.group(2)
        return molecule(formula, coefficient=coef, phase=phase)  # list of molecule objects

    def _parse_input_str(self,s):
        compound_list = s.split(' + ')
        return [self._parse_compound_str(c) for c in compound_list]
        
        
    def _generate_rxn_plain_text(self):
        def _get_rxn_term(m):
            coef_str = f'{m.coefficient}' if m.coefficient > 1 else ''
            phase_str = f'({m.phase})' if m.phase is not None else ''
            return coef_str+str(m)+phase_str
        
        reactant_terms = [_get_rxn_term(m) for m in self.reactants]
        product_terms = [_get_rxn_term(m) for m in self.products]
        
        plaintext = f'{" + ".join(reactant_terms)}'+' = '+f'{" + ".join(product_terms)}'
        
        reactant_perm = list(permutations(reactant_terms))
        product_perm = list(permutations(product_terms))
        all_combinations = [(r, p) for r in reactant_perm for p in product_perm]
        all_answer = [f'{" + ".join(r)}'+' = '+f'{" + ".join(p)}' for r,p in all_combinations]
        plaintext_answers=';'.join(all_answer)

        return plaintext,plaintext_answers
    
    def _generate_simple_html(self):
        def _get_rxn_term(m):
            coef_str = f'{m.coefficient}' if m.coefficient > 1 else ''
            phase_str = f'({m.phase})' if m.phase is not None else ''
            return coef_str + m.simple_html + phase_str
        
        reactant_terms = [_get_rxn_term(m) for m in self.reactants]
        product_terms = [_get_rxn_term(m) for m in self.products]
        
        simple_html = f'{" + ".join(reactant_terms)}'+' &rarr; '+f'{" + ".join(product_terms)}'
        return simple_html
   
    def _generate_rxn_tex(self):
        reactant_terms = [f'{m.coefficient if m.coefficient != 1 else ""}{m.tex}{"("+m.phase+")" if m.phase else ""}' for m in self.reactants]
        product_terms = [f'{m.coefficient if m.coefficient != 1 else ""}{m.tex}{"("+m.phase+")" if m.phase else ""}' for m in self.products]
        
        arrow = '\\leftrightarrows{}' if self.reversible is True else '\\rightarrow{}'

        tex = f'{"+".join(reactant_terms)}'+arrow+f'{"+".join(product_terms)}'
        return tex
    
    def _check_balance(self):
        
        charge_reactants = sum([m.charge for m in self.reactants])
        charge_products = sum([m.charge for m in self.products])

        # Cumulative element dictionary for reactants
        reactant_elements = {}
        for m in self.reactants:
            for elem, count in m.element_dict.items():
                if elem == 'e-':  # Skip if the element is 'e-'
                    continue
                if elem in reactant_elements:
                    reactant_elements[elem] += count*m.coefficient
                else:
                    reactant_elements[elem] = count*m.coefficient

        # Cumulative element dictionary for products
        product_elements = {}
        for m in self.products:
            for elem, count in m.element_dict.items():
                if elem == 'e-':  # Skip if the element is 'e-'
                    continue
                if elem in product_elements:
                    product_elements[elem] += count*m.coefficient
                else:
                    product_elements[elem] = count*m.coefficient

        # Check if the two dictionaries match and charges are the same
        if reactant_elements == product_elements and charge_reactants == charge_products:
            return True
        else:
            return False

    def _generate_rxn_quotient(self):
        reactants = [m for m in self.reactants if (str(m) != 'e-' and m.phase is None)]
        products = [m for m in self.products if (str(m) != 'e-' and m.phase is None)]
        
        reactant_terms = [f'[{m}]^{m.coefficient}' if m.coefficient > 1 else f'[{m}]' for m in reactants]
        product_terms = [f'[{m}]^{m.coefficient}' if m.coefficient > 1 else f'[{m}]' for m in products]

        Q_eqn = f'{"".join(product_terms)}'+('/' if len(reactant_terms)>0 and len(product_terms)>0 else '')+f'{"".join(reactant_terms)}'

        reactant_perm = list(permutations(reactant_terms))
        product_perm = list(permutations(product_terms))

        all_combinations = [(r, p) for r in reactant_perm for p in product_perm]
        all_Q = [f'{"".join(p)}'+('/' if len(r)>0 and len(p)>0 else '')+f'{"".join(r)}' for r,p in all_combinations]
        Q_eqn_answers=';'.join(all_Q)

        reactant_terms_tex = [f'[{m.tex}]^{{{m.coefficient}}}' if m.coefficient > 1 else f'[{m.tex}]' for m in reactants]
        product_terms_tex = [f'[{m.tex}]^{{{m.coefficient}}}' if m.coefficient > 1 else f'[{m.tex}]' for m in products]
        Q_tex = f'\\frac{{{"".join(product_terms_tex)}}}{{{"".join(reactant_terms_tex)}}}'


        if all(m.concentration is not None for m in reactants) and all(m.concentration is not None for m in products):
            Q=1
            for r in reactants:
                Q *= r.concentration**(-1*r.coefficient)
            for p in products:
                Q *= p.concentration**(p.coefficient)
        else:
            Q = None

        return Q_eqn, Q_eqn_answers, Q_tex, Q
    
    def _generate_rate_expression(self):
        reactants = [m for m in self.reactants if (str(m) != 'e-' and m.phase is None)]
        
        reactant_terms = [f'[{m}]^{m.coefficient}' if m.coefficient > 1 else f'[{m}]' for m in reactants]

        rate_eqn = 'k'+f'{"".join(reactant_terms)}'

        reactant_perm = list(permutations(reactant_terms))
        
        all_eqn = ['k'+f'{"".join(r)}' for r in reactant_perm]
        rate_eqn_answers=';'.join(all_eqn)

        reactant_terms_tex = [f'[{m.tex}]^{{{m.coefficient}}}' if m.coefficient > 1 else f'[{m.tex}]' for m in reactants]
        rate_eqn_tex = f'k{"".join(reactant_terms_tex)}'


        if all(m.concentration is not None for m in reactants) and self.k is not None:
            rate=self.k
            for r in reactants:
                rate *= r.concentration**(r.coefficient)
        else:
            rate = None

        return rate_eqn, rate_eqn_answers, rate_eqn_tex, rate

    def set_concentration(self, s, conc):
        """
        s is the plaintext representation of the species whose concentration is to be set. e.g. 'SO4--' for sulfate
        conc is the concentration, entered as a number

        this function automatically calls the updater, ensuring that the Q value is calculated once all concentrations are defined
        """
        
        # Search for the molecule in reactants and set the concentration
        for m in self.reactants:
            if str(m) == s:
                m.concentration = conc
                self._update()
                return

        # Search for the molecule in products and set the concentration
        for m in self.products:
            if str(m) == s:
                m.concentration = conc
                self._update()
                return

        # If the molecule was not found in either list, you can optionally raise an error or print a message
        print(f"Molecule {s} not found in reactants or products.")



import glob
import xarray as xr
import numpy as np
from datetime import datetime, timedelta
from zoneinfo import ZoneInfo
from scipy.spatial import cKDTree

from .timeseries import timeseries

class camxdata():
    def __init__(self, dir = None):
        if dir is None:
            raise TypeError("interactive file selector is not written yet")
        else:
            self.files=sorted(glob.glob(dir+"/*.nc"))
            self.dir=dir

        #initialize times, lat, lon, etc, and validate each file as it gets checked
        file_times = []
        file_starttime = []
        file_stoptime = []
        times = []
         
        for i, file in enumerate(self.files):
            ds = xr.open_dataset(file)
            if i == 0:
                self.latitude = ds.latitude
                self.longitude = ds.longitude
                self.topo = ds.topo
                self.z = ds.z
                self.LAY = len(ds['LAY'])
                self.ROW = len(ds['ROW'])
                self.COL = len(ds['COL'])
                self.variables = list(ds.data_vars)
                
            f_lat = ds.latitude
            f_lon = ds.longitude
            f_topo = ds.topo
            f_lay = len(ds['LAY'])

            if not self.latitude.equals(f_lat) or not self.longitude.equals(f_lon) or not self.topo.equals(f_topo) or f_lay != self.LAY:
                raise IOError(f".nc files in the file list are not congruent in space. error occured reading {file}")

            for v in list(ds.data_vars):
                if v not in self.variables:
                    raise ValueError(f"files in the file list have different variables. error occured reading {file}")

            YYYYjjj = ds['TFLAG'][:,0,0].values #Year/day-of-year format
            hhmmss = ds['TFLAG'][:,0,1].values
        
            # get starttimes for each point
            dt = [(datetime.strptime(str(a)+"+0000","%Y%j%z") + timedelta(hours=b/10000)) for a,b in zip(YYYYjjj,hhmmss)]
            file_times.append(np.array(dt))
            times.extend(dt)

            #get file start and stop to enable lookups later
            time_step_hr = ds.attrs['TSTEP']/10000
            file_starttime.append(dt[0])
            file_stoptime.append(dt[-1]+timedelta(hours=time_step_hr))
        
        #get time metadata set up
        self.file_times=np.array(file_times)
        self.file_starttime=np.array(file_starttime)
        self.file_stoptime=np.array(file_stoptime)
        self.file_ts = timeseries(self.file_starttime,np.arange(len(self.files)),stop_times=self.file_stoptime)
        self.time=np.array(times)
        self.stoptime = max(self.time)
        self.starttime = min(self.time)

        #build tree for nearest neighbors searching (2D)
        lat_flat = self.latitude.values.ravel()
        lon_flat = self.longitude.values.ravel()
        self.tree = cKDTree(np.column_stack([lat_flat, lon_flat]))


        

    def _get_time_file_index(self,dt_utc: "datetime"):
        """Take a datetime, return index of file and index of timestamp as a tuple"""
        ifile = self.file_ts.instantaneous_value(dt_utc)
        times = self.file_times[ifile]
        itime = np.searchsorted(times,dt_utc,side="right")-1
        return ifile,itime
    
    def _dt_to_utc(self,dt):
        if dt.tzinfo is None:
            raise ValueError("No timezone-naive timestamps!")
        return dt.astimezone(ZoneInfo("UTC"))
    
    def _get_row_col(self,lat,lon):
        _, idx = self.tree.query([lat, lon]) # use ckdtree
        row, col = np.unravel_index(idx, self.latitude.shape) # Map flat index back to 2D row/col
        return row, col
    
    def _get_layer(self,z_m_agl,lat,lon,time,AMSL=False):
        """
        This gets the layer index for a given altitude/lat/lon at a fixed point in time (since layer heights vary in time)
        """
        dt_utc = self._dt_to_utc(time)
        FILE, TSTEP=self._get_time_file_index(dt_utc)       
        ROW, COL = self._get_row_col(lat,lon)

        topo_z = self.topo[ROW][COL].values
        if AMSL:    #if input given relative to sea level, subtract off ground level to get AGL
            z_m_agl -= topo_z

        ds = xr.open_dataset(self.files[FILE])
        z_vals = ds.z.isel(TSTEP=TSTEP, ROW=ROW, COL=COL).values
        lay = np.abs(z_vals - z_m_agl).argmin()
        return lay
            
    def get(self, var, time=None, LAT=None, LON=None, z_m_agl = None):
        """
        infers what kind of dataset you're asking for based on the inputs you give
        ...probably going to be a shitshow for a while...
        """

        if var not in self.variables:
            raise ValueError(f"{var} not in variables of this camxdata instance")

        

        if time is None:
            time = self.time

        if LAT is not None and LON is not None: #we have defined space
            if len(LAT) > 1 and len(LAT) == len(LON):
                
                if isinstance(time,datetime):
                    raise ValueError('camxdata.get cant do this yet (LAT/LON arrays, single timestamp)')
                
                if len(time) > 1:   #Multiple times in multiple spaces
                    #return a timeseries object (usos.timeseries) with LAT/LON data encoded

                    print("Lydia's code goes here") 
                    print(f"DEBUG: len(time) = {len(time)}, len(LAT) = {len(LAT)}, len(LON) = {len(LON)}")

                    if not (len(time) == len(LAT) == len(LON)):
                        raise ValueError("For multiple points/times, time, LAT, and LON must all be the same length")

                    values = []
                    times_out = []
                    open_file = None
                    ds = None
                    n_total = 0
                    n_appended = 0

                    try:
                        for dt, lat_i, lon_i in zip(time, LAT, LON):
                            n_total += 1
                            dt_utc = self._dt_to_utc(dt)

                            # ALWAYS append something so outputs align 1:1 with inputs
                            if dt_utc < self.starttime or dt_utc > self.stoptime:
                                values.append(np.nan)
                                times_out.append(dt)
                                n_appended += 1
                                continue

                            ifile, TSTEP = self._get_time_file_index(dt_utc)
                            if ifile != open_file:
                                if ds is not None:
                                    ds.close()
                                ds = xr.open_dataset(self.files[ifile])
                                open_file = ifile

                            ROW, COL = self._get_row_col(lat_i, lon_i)
                            if z_m_agl is None:
                                LAY = 0
                            else:
                                # Reuse open ds and known TSTEP to avoid double-open
                                LAY = self._get_layer(z_m_agl, lat_i, lon_i, dt, ds=ds, TSTEP=TSTEP)

                            # Robust, name-based indexing (handles ROW/COL vs Y/X)
                            sel = {}
                            for name in ds[var].dims:
                                u = name.upper()
                                if u == 'TSTEP':
                                    sel[name] = TSTEP
                                elif u == 'LAY':
                                    sel[name] = LAY
                                elif u in ('ROW', 'Y'):
                                    sel[name] = ROW
                                elif u in ('COL', 'X'):
                                    sel[name] = COL

                            try:
                                v = ds[var].isel(**sel).values
                                try:
                                    v = float(v)
                                except Exception:
                                    arr = np.asarray(v)
                                    v = arr.item() if arr.size == 1 else np.nan
                            except Exception as e:
                                # If indexing fails for any reason, append NaN but keep alignment
                                # (and you can log the first few errors if needed)
                                v = np.nan

                            values.append(v)
                            times_out.append(dt)
                            n_appended += 1
                    finally:
                        if ds is not None:
                            ds.close()

                    print(f"[camxdata.get route] iterated={n_total} appended={n_appended}")
                    return timeseries(times_out, values)  # don't rely on .LAT/.LON on the timeseries
                                
            else: # single point in space

                ROW,COL = self._get_row_col(LAT,LON)

                if isinstance(time,datetime): #single point in space and time
                    
                    #get file, time
                    dt_utc = self._dt_to_utc(time)
                    FILE, TSTEP=self._get_time_file_index(dt_utc)
                    ds = xr.open_dataset(self.files[FILE])

                    # get layer
                    if z_m_agl is None:
                        LAY = 0 #default to surface layer
                    else:
                        LAY = self._get_layer(z_m_agl,LAT,LON,time)

                    #get value
                    val = ds[var][TSTEP][LAY][ROW][COL].values
                    ds.close()
                    return val
                
                elif len(time) > 1: #single point in space, multiple points in time
                    
                    # map out points in time to a list of files. then loop through files, grabbing all points from that file
                    # before closing
                    values = []
                    times = []
                    open_file = None
                    for dt in time:
                        dt_utc = self._dt_to_utc(dt)
                        
                        if dt_utc < self.starttime or dt_utc > self.stoptime:
                            continue
                    
                        ifile, TSTEP = self._get_time_file_index(dt_utc)
                        if ifile != open_file:
                            if open_file is not None:
                                ds.close()
                            ds = xr.open_dataset(self.files[ifile])
                            open_file = ifile

                        # get layer
                        if z_m_agl is None:
                            LAY = 0 #default to surface layer
                        else:
                            LAY = self._get_layer(z_m_agl,LAT,LON,time)   #RIGHT NOW THIS DOUBLE-OPENS THE DS. MODIFY TO ACCCEPT THE Z FIELD AS AN INPUT
                        
                        val = ds[var][TSTEP][LAY][ROW][COL].values
                        values.append(val)
                        times.append(dt)

                    return timeseries(times,values)
        raise ValueError("That's not a supported combination of inputs for camxdata.get()")

    def average_timeseries_surface(self, ts : timeseries):
        if ts.LAT is None or ts.LON is None:
            raise ValueError('ts passed to camxdata.average_timeseries_surface is missing LAT/LON data')

        valid_mask = np.isfinite(ts.LAT) & np.isfinite(ts.LON)
        valid_lat = ts.LAT[valid_mask]
        valid_lon = ts.LON[valid_mask]
        valid_data = ts.data[valid_mask] 

        # Query nearest grid point for each data point
        _, idx = self.tree.query(np.column_stack([valid_lat, valid_lon]))

        # Accumulate statistics
        bin_dict = {}

        for i, val in zip(idx, valid_data):
            if i not in bin_dict:
                bin_dict[i] = []
            bin_dict[i].append(val)


        stats_arrays = {key: np.full((self.ROW,self.COL), np.nan) for key in ['mean', 'median', 'std', 'min', 'max', 'N']}
        
        for i, values in bin_dict.items():
            y, x = np.unravel_index(i, (self.ROW,self.COL))
            arr = np.array(values)
            N = np.sum(~np.isnan(arr))
            if N>0:
                stats_arrays['mean'][y, x] = np.nanmean(arr)
                stats_arrays['median'][y, x] = np.nanmedian(arr)
                stats_arrays['std'][y, x] = np.nanstd(arr)
                stats_arrays['min'][y, x] = np.nanmin(arr)
                stats_arrays['max'][y, x] = np.nanmax(arr)
                stats_arrays['N'][y, x] = N

        # Create Dataset
        ds = xr.Dataset(
            {key: (("y", "x"), val) for key, val in stats_arrays.items()},
            coords={
                "latitude": (("y", "x"), self.latitude.data),
                "longitude": (("y", "x"), self.longitude.data)
            }
        )

        return ds
    

    def get_surface_field(self, var, time):
        dt_utc = self._dt_to_utc(time)
        FILE, TSTEP = self._get_time_file_index(dt_utc)
        LAY= 0
        ds = xr.open_dataset(self.files[FILE])
        field = ds[var][TSTEP][LAY]
        return field



import os
import inspect

def save(fig):
    # Find the outermost script (__main__)
    for frame in inspect.stack():
        module = inspect.getmodule(frame.frame)
        if module and module.__name__ == "__main__":
            parent_script = module.__file__
            break
    else:
        raise RuntimeError("Could not determine the outermost script (__main__).")

    # Construct the figure path
    parent_dir = os.path.dirname(os.path.abspath(parent_script))
    script_name = os.path.basename(parent_script)
    fig_name = script_name.replace('.py', '.png')
    fig_path = os.path.join(parent_dir, fig_name)

    # Save the figure
    fig.savefig(fig_path)
    print(f"Figure saved to: {fig_path}")
    return